/* 
 * Datei: pcf8583.c
 * Info: Routinen zum Einlesen und Setzen der i2C-Uhr PCF8583
 * Autor: Werner Bllmann
 * Datum: 10.04.2005
 * 21.4.2005 wb - sommerzeitstatus in RAM 0x12
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <compat/twi.h>

#include "mcu.h"
#include "pcf8583.h"
#include "tools.h"

#define PCF8583ADDR  (0x50)

/*
    SCLfrq = F_CPU / (16 + 2*(TWBR)*4^TWPS) 
    Mit TWPS == 0 wird 4^TWPS = 1
      SCLfrq = F_PCU / (16 + 2*TWBR)
    fr SCLfrq == 100kHz
      100000*(16 + 2*TWBR) = F_CPU
      1600000 + 200000*TWBR = F_CPU
      200000*TWBR = F_CPU - 1600000
      TWBR =  (F_CPU-1600000)/200000
      TWBR =  (F_CPU/100000 - 16) / 2
    Nach Dateblatt soll TWBR nicht kleiner als 10 werden.
    Ausgehend von F_CPU_MAX = 20MHz wird TWBRmax = 92 bei 100kHz TWI
    Fuer 400kHz TWI waehren das 368 -> Anderer Prescaler noetig.
*/
#if F_CPU < 3600000UL
  #define TWBR_VAL  10                    /* smallest TWBR value, see note in the datasheet */
  /*
    The datasheet explains why a minimum TWBR value of 10 should be maintained when running 
    in master mode. Thus, for system clocks below 3.6 MHz, we cannot run the bus at the 
    intented clock rate of 100 kHz but have to slow down accordingly.
  */  
#else
  #define TWBR_VAL  ((F_CPU / 100000UL - 16) / 2)
#endif

/*
 * Saved TWI status register, for error messages only.  We need to
 * save it in a variable, since the datasheet only guarantees the TWSR
 * register to have valid contents while the TWINT bit in TWCR is set.
 */
uint8_t twst = 0;

void twi_init(void)
{
/* initialize TWI clock: 100 kHz clock, TWPS = 0 => prescaler = 1 */
#if defined(TWPS0)
    /* has prescaler (mega128 & newer) */
#if defined (__AVR_ATmega32__) || defined (__AVR_ATmega8535__)
    DDRC &= ~(_BV(PC1)|_BV(PC0));
#elif defined (__AVR_ATmega128__)
    DDRD &= ~(_BV(PD1)|_BV(PD0));
#elif defined (__AVR_ATmega8__)
    DDRC &= ~(_BV(PC5)|_BV(PC4));
#else
        #error TWI Port pins not defined for this MCU yet!
#endif    
    TWSR = 0;  // prescaler = 1
#endif
    TWBR = TWBR_VAL;
}

/*
 * 
 */
/* 
SIGNAL (SIG_2WIRE_SERIAL)
{
}
*/ 
/*
 * Read "len" bytes from device starting at "addr" into "buf".
 *
 * This requires two bus cycles: during the first cycle, the device
 * will be selected (master transmitter mode), and the address
 * transfered.  
 *
 * The second bus cycle will reselect the device (repeated start
 * condition, going into master receiver mode), and transfer the data
 * from the device to the TWI master.  Multiple bytes can be
 * transfered by ACKing the client's transfer.  The last transfer will
 * be NACKed, which the client will take as an indication to not
 * initiate further transfers.
 */
int8_t twi_read_bytes(uint8_t sla, uint8_t addr, uint8_t len, uint8_t *buf)
{
  uint8_t twcr, n = 0;
  int8_t rv = 0;

  sla <<= 1;  // make a write address
  
  /*
   * Note [8]
   * First cycle: master transmitter mode
   */
  restart:
  if (n++ >= MAX_ITER)
    return -1;
  begin:

  TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN); /* send start condition */
  while (!(TWCR & _BV(TWINT))) /* wait for transmission */
     ; 
  switch ((twst = TW_STATUS))
  {
    case TW_REP_START:          /* OK, but should not happen */
    case TW_START:
      break;

    case TW_MT_ARB_LOST:        /* Note [9] */
      goto begin;

    default:
      return -1;                /* error: not in start condition */
                                /* NB: do /not/ send stop condition */
  }

  /* Note [10] */
  /* send SLA+W */
  TWDR = sla; // is 0 // | TW_WRITE;
  TWCR = _BV(TWINT) | _BV(TWEN);      /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
      ; 
  switch ((twst = TW_STATUS))
  {
    case TW_MT_SLA_ACK:
      break;

    case TW_MT_SLA_NACK:          /* nack during select: device busy writing */
                                  /* Note [11] */
      goto restart;

    case TW_MT_ARB_LOST:          /* re-arbitrate */
      goto begin;

    default:
      goto error;                 /* must send stop condition */
  }

  TWDR = addr;                    /* 8 bits of addr */
  TWCR = _BV(TWINT) | _BV(TWEN); /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
    ; 
  switch ((twst = TW_STATUS))
  {
    case TW_MT_DATA_ACK:
      break;

    case TW_MT_DATA_NACK:
      goto quit;

    case TW_MT_ARB_LOST:
      goto begin;

    default:
      goto error;               /* must send stop condition */
  }

  /*
   * Note [12]
   * Next cycle(s): master receiver mode
   */
  TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN); /* send (rep.) start condition */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
    ; 
  switch ((twst = TW_STATUS))
  {
    case TW_START:              /* OK, but should not happen */
    case TW_REP_START:
      break;

    case TW_MT_ARB_LOST:
      goto begin;

    default:
      goto error;
  }

  /* send SLA+R */
  TWDR = sla | TW_READ;
  TWCR = _BV(TWINT) | _BV(TWEN); /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
     ; 
  switch ((twst = TW_STATUS))
  {
    case TW_MR_SLA_ACK:
      break;

    case TW_MR_SLA_NACK:
      goto quit;

    case TW_MR_ARB_LOST:
      goto begin;

    default:
      goto error;
  }

  for (twcr = _BV(TWINT) | _BV(TWEN) | _BV(TWEA) /* Note [13] */;
       len != 0;
       len--)
  {
      if (len == 1)
        twcr = _BV(TWINT) | _BV(TWEN); /* send NAK this time */
      TWCR = twcr;              /* clear int to start transmission */

      while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
          ; 
      switch ((twst = TW_STATUS))
      {
        case TW_MR_DATA_NACK:
          len = 1;              /* force end of loop */
          /* FALLTHROUGH */
        case TW_MR_DATA_ACK:
          *buf++ = TWDR;
          rv++;
          break;

        default:
          goto error;
      }
  }
  quit:
/* Note [14] */
  TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN); /* send stop condition */

  return rv;

  error:
  rv = -1;
  goto quit;
}

/*
 * Write "len" bytes into device starting at "addr" from "buf".
 *
 * This is a bit simpler than the previous function since both, the
 * address and the data bytes will be transfered in master transmitter
 * mode, thus no reselection of the device is necessary.  However, the
 * EEPROMs are only capable of writing one "page" simultaneously, so
 * care must be taken to not cross a page boundary within one write
 * cycle.  The amount of data one page consists of varies from
 * manufacturer to manufacturer: some vendors only use 8-byte pages
 * for the smaller devices, and 16-byte pages for the larger devices,
 * while other vendors generally use 16-byte pages.  We thus use the
 * smallest common denominator of 8 bytes per page, declared by the
 * macro PAGE_SIZE above.
 *
 * The function simply returns after writing one page, returning the
 * actual number of data byte written.  It is up to the caller to
 * re-invoke it in order to write further data.
 */
int8_t
twi_write_bytes(uint8_t sla, uint8_t addr, uint8_t len, uint8_t *buf)
{
  uint8_t n = 0;
  int8_t rv = 0;
  uint8_t endaddr;

  endaddr = addr + len;
  if(endaddr < addr)
    return -1;

  sla <<= 1;

  restart:
  if (n++ >= MAX_ITER)
    return -1;
  begin:

  /* Note [15] */
  TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN); /* send start condition */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
    ; 
  switch ((twst = TW_STATUS))
  {
    case TW_REP_START:          /* OK, but should not happen */
    case TW_START:
      break;

    case TW_MT_ARB_LOST:
      goto begin;

    default:
      return -1;                /* error: not in start condition */
                                /* NB: do /not/ send stop condition */
  }

  /* send SLA+W */
  TWDR = sla; //  | TW_WRITE;
  TWCR = _BV(TWINT) | _BV(TWEN); /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
    ; 
  switch ((twst = TW_STATUS))
  {
    case TW_MT_SLA_ACK:
      break;

    case TW_MT_SLA_NACK:        /* nack during select: device busy writing */
      goto restart;

    case TW_MT_ARB_LOST:        /* re-arbitrate */
      goto begin;

    default:
      goto error;               /* must send stop condition */
  }

  TWDR = addr;                  /* 8 bits of addr */
  TWCR = _BV(TWINT) | _BV(TWEN); /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
    ; 
  switch ((twst = TW_STATUS))
  {
    case TW_MT_DATA_ACK:
      break;

    case TW_MT_DATA_NACK:
      goto quit;

    case TW_MT_ARB_LOST:
      goto begin;

    default:
      goto error;               /* must send stop condition */
  }

  for (; len > 0; len--)
  {
      TWDR = *buf++;
      TWCR = _BV(TWINT) | _BV(TWEN); /* start transmission */
      while ((TWCR & _BV(TWINT)) == 0) /* wait for transmission */
        ; 
      switch ((twst = TW_STATUS))
      {
        case TW_MT_DATA_NACK:
          goto error;           /* device write protected -- Note [16] */

        case TW_MT_DATA_ACK:
          rv++;
          break;

        default:
          goto error;
      }
  }
  quit:
  TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN); /* send stop condition */

  return rv;

  error:
  rv = -1;
  goto quit;
}

int8_t read_pcf8583(pcf8583_ctl *data)
{
    int8_t rv;
    
    // Lese volles Jahr + SommerzeitStatus aus CMOS-RAM
    if((rv = twi_read_bytes(PCF8583ADDR, FULL_YEAR_ADDR, FULL_YEAR_LEN+1, (uint8_t *)&data->year)) == -1)
    {
//        TWCR = _BV(TWINT);
//        TWCR = _BV(TWINT);// | _BV(TWSTO) | _BV(TWEN); /* send stop condition */
//        return -1;
    }
    data->control = CAPTURE;
    if((rv = twi_read_bytes(PCF8583ADDR, 0, 8, &data->bytes[0])) == -1)
    {
//        TWCR = _BV(TWINT);
//        TWCR = _BV(TWINT);// | _BV(TWSTO) | _BV(TWEN); /* send stop condition */
//        return -1;
    }
    data->control = 0;
    twi_write_bytes(PCF8583ADDR, 0, 1, (uint8_t *)&data->bytes[0]);
    data->bytes[1] = fromBCD(data->bytes[1]); // h_sec 
    data->bytes[2] = fromBCD(data->bytes[2]); // seconds
    data->bytes[3] = fromBCD(data->bytes[3]); // minutes
    data->bytes[4] = fromBCD(data->bytes[4] & 0x3F); // hours
    data->days     = fromBCD((uint8_t)data->days);
    data->month    = fromBCD((uint8_t)data->month);
    
    // Falls das Jahr weitergezhlt hat, nachholen und merken.
    //  So kann die Hardware 3 Jahre ausgeschaltet bleiben ;)  
    if((data->year & 3) != data->yr)
    {
        while((data->year & 3) != data->yr)
            data->year++;
        twi_write_bytes(PCF8583ADDR, FULL_YEAR_ADDR, FULL_YEAR_LEN, (uint8_t *)&data->year);
    } 
    TWCR = _BV(TWINT);// | _BV(TWSTO) | _BV(TWEN); /* send stop condition */
    TWCR = _BV(TWINT);
    return rv;
}


int8_t write_pcf8583(pcf8583_ctl *data)
{
    uint8_t  dummy;
    data->control = STOP | CAPTURE;
    data->h_sec   = toBCD(data->h_sec);
    data->sec     = toBCD(data->sec);
    data->min     = toBCD(data->min);
    data->hr      = toBCD(data->hr & 0x1F);
    data->days    = toBCD((uint8_t)data->days);
    data->month   = toBCD((uint8_t)data->month);
    data->yr      = data->year;     // Unterste 2 bit Jahr nach RTC
//    data->weekday = 0;  
    
    // Jahr+Sommerzeitstatus in RAM ab offset FULL_YEAR_ADDR schreiben
    if( twi_write_bytes(PCF8583ADDR, FULL_YEAR_ADDR, FULL_YEAR_LEN+1, (uint8_t *)&data->year) == -1)
    {
//        TWCR = _BV(TWINT);// | _BV(TWSTO) | _BV(TWEN); /* send stop condition */
//        return -1;
    }
    // Restliches Datum/Zeit schreiben. Ignorieren h_sec
    twi_write_bytes(PCF8583ADDR, 0, 8, &data->bytes[0]);

    dummy = 0; //data->control & ~(STOP|CAPTURE);
    dummy = twi_write_bytes(PCF8583ADDR, 0, 1, &dummy);
    TWCR = _BV(TWINT);// | _BV(TWSTO) | _BV(TWEN); /* send stop condition */
    TWCR = _BV(TWINT);
    return dummy;
}
